#ifndef __CL_RAW_LIST_H
#define __CL_RAW_LIST_H

/*___INFO__MARK_BEGIN__*/
/*************************************************************************
 *
 *  The Contents of this file are made available subject to the terms of
 *  the Sun Industry Standards Source License Version 1.2
 *
 *  Sun Microsystems Inc., March, 2001
 *
 *
 *  Sun Industry Standards Source License Version 1.2
 *  =================================================
 *  The contents of this file are subject to the Sun Industry Standards
 *  Source License Version 1.2 (the "License"); You may not use this file
 *  except in compliance with the License. You may obtain a copy of the
 *  License at http://gridengine.sunsource.net/Gridengine_SISSL_license.html
 *
 *  Software provided under this License is provided on an "AS IS" basis,
 *  WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING,
 *  WITHOUT LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS,
 *  MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE, OR NON-INFRINGING.
 *  See the License for the specific provisions governing your rights and
 *  obligations concerning the Software.
 *
 *  The Initial Developer of the Original Code is: Sun Microsystems, Inc.
 *
 *  Copyright: 2001 by Sun Microsystems, Inc.
 *
 *  All Rights Reserved.
 *
 ************************************************************************/
/*___INFO__MARK_END__*/


int                 cl_raw_list_setup                   (cl_raw_list_t** list_p, char* list_name, int enable_list_locking);         /* CR check */
int                 cl_raw_list_cleanup                 (cl_raw_list_t** list_p);                                 /* CR check */
cl_raw_list_elem_t* cl_raw_list_append_elem             (cl_raw_list_t* list_p, void* data);                      /* CR check */
void*               cl_raw_list_remove_elem             (cl_raw_list_t* list_p, cl_raw_list_elem_t* delete_elem); /* CR check */
int                 cl_raw_list_dechain_elem            (cl_raw_list_t* list_p, cl_raw_list_elem_t* dechain_elem);
int                 cl_raw_list_append_dechained_elem   (cl_raw_list_t* list_p, cl_raw_list_elem_t* dechain_elem);
int                 cl_raw_list_lock                    (cl_raw_list_t* list_p);                                  /* CR check */
int                 cl_raw_list_unlock                  (cl_raw_list_t* list_p);                                  /* CR check */
unsigned long       cl_raw_list_get_elem_count          (cl_raw_list_t* list_p);                                  /* CR check */
cl_raw_list_elem_t* cl_raw_list_search_elem             (cl_raw_list_t* list_p, void* data);                      /* CR check */
cl_raw_list_elem_t* cl_raw_list_get_first_elem          (cl_raw_list_t* list_p);                                  /* CR check */
cl_raw_list_elem_t* cl_raw_list_get_least_elem          (cl_raw_list_t* list_p);
cl_raw_list_elem_t* cl_raw_list_get_next_elem           (cl_raw_list_elem_t* elem);                               /* CR check */
cl_raw_list_elem_t* cl_raw_list_get_last_elem           (cl_raw_list_elem_t* elem);                               /* CR check */

#endif /* __CL_RAW_LIST_H */

